from collections import namedtuple
import threading

import psycopg2
import psycopg2.extras
import psycopg2.extensions
import psycopg2.errorcodes

from .log import logger

psycopg2.extensions.register_type(psycopg2.extensions.UNICODE, None)
psycopg2.extensions.register_type(psycopg2.extensions.UNICODEARRAY, None)

PrimaryKeyMapItem = namedtuple('PrimaryKeyMapItem', 'table_name, col_name, col_type, col_ord_pos')


class SlotReader(object):
    PK_SQL = """
    SELECT CONCAT(table_schema, '.', table_name), column_name, data_type, ordinal_position
    FROM information_schema.tables
    LEFT JOIN (
        SELECT CONCAT(table_schema, '.', table_name), column_name, data_type, c.ordinal_position,
                    table_catalog, table_schema, table_name
        FROM information_schema.table_constraints
        JOIN information_schema.key_column_usage AS kcu
            USING (constraint_catalog, constraint_schema, constraint_name,
                        table_catalog, table_schema, table_name)
        JOIN information_schema.columns AS c
            USING (table_catalog, table_schema, table_name, column_name)
        WHERE constraint_type = 'PRIMARY KEY'
    ) as q using (table_catalog, table_schema, table_name)
    ORDER BY ordinal_position;
    """

    def __init__(self, database, host, port, user, slot_name,
                 output_plugin='test_decoding', keepalive_window=30):
        # Cool fact: using connections as context manager doesn't close them on
        # success after leaving with block
        self._db_confg = dict(database=database, host=host, port=port, user=user)
        self._keepalive_window = keepalive_window
        self._repl_conn = None
        self._repl_cursor = None
        self._normal_conn = None
        self.slot_name = slot_name
        self.output_plugin = output_plugin
        self.cur_lag = 0

    def __enter__(self):
        self._normal_conn = self._get_connection()
        self._normal_conn.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        self._repl_conn = self._get_connection(connection_factory=psycopg2.extras.LogicalReplicationConnection)
        self._repl_cursor = self._repl_conn.cursor()

        if self._keepalive_window:
            self._keepalive_thread = self._send_keepalive()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Be a good citezen and try to clean up on the way out.
        """
        if self._keepalive_thread:
            try:
                self._keepalive_thread.join(timeout=self._keepalive_window+3)
            except Exception:
                pass

        try:
            self._repl_cursor.close()
        except Exception:
            pass

        try:
            self._repl_conn.close()
        except Exception:
            pass

        try:
            self._normal_conn.close()
        except Exception:
            pass


    def _send_keepalive(self):

        # keepalive with no args.
        try:
            self._repl_cursor.send_feedback()
        except psycopg2.DatabaseError as e:
            if not e.message == 'no COPY in progress\n':
                logger.exception(e)
        except Exception as e:
            logger.exception(e)

        # schedule myself in the future
        t = threading.Timer(self._keepalive_window, self._send_keepalive)
        t.daemon = True
        t.start()
        return t

    def _get_connection(self, connection_factory=None, cursor_factory=None):
        return psycopg2.connect(connection_factory=connection_factory,
                                cursor_factory=cursor_factory, **self._db_confg)

    def _execute_and_fetch(self, sql, *params):
        with self._normal_conn.cursor() as cur:
            if params:
                cur.execute(sql, params)
            else:
                cur.execute(sql)

            return cur.fetchall()

    @property
    def primary_key_map(self):
        logger.info('Getting primary key map')
        result = map(PrimaryKeyMapItem._make, self._execute_and_fetch(SlotReader.PK_SQL))
        pk_map = {rec.table_name: rec for rec in result}

        return pk_map

    def create_slot(self):
        logger.info('Creating slot %s' % self.slot_name)
        try:
            self._repl_cursor.create_replication_slot(self.slot_name,
                                                      slot_type=psycopg2.extras.REPLICATION_LOGICAL,
                                                      output_plugin=self.output_plugin)
        except psycopg2.ProgrammingError as p:
            # Will be raised if slot exists already.
            if p.pgcode != psycopg2.errorcodes.DUPLICATE_OBJECT:
                logger.error(p)
                raise
            else:
                logger.info('Slot %s is already present.' % self.slot_name)

    def delete_slot(self):
        logger.info('Deleting slot %s' % self.slot_name)
        try:
            self._repl_cursor.drop_replication_slot(self.slot_name)
        except psycopg2.ProgrammingError as p:
            # Will be raised if slot exists already.
            if p.pgcode != psycopg2.errorcodes.UNDEFINED_OBJECT:
                logger.error(p)
                raise
            else:
                logger.info('Slot %s was not found.' % self.slot_name)

    def process_replication_stream(self, consume):
        logger.info('Starting the consumption of slot "%s"!' % self.slot_name)
        if self.output_plugin == 'wal2json':
            options = {'include-xids': 1}
        else:
            options = None
        self._repl_cursor.start_replication(self.slot_name, options=options)
        self._repl_cursor.consume_stream(consume)



__version__ = '0.6.0'

__uri__ = 'https://github.com/handshake/pg2kinesis'
__title__ = 'pg2kinesis'
__description__ = 'A utility that enables PostgreSQL 9.4+ to logically replicate to Amazon''s Kinesis'
__doc__ = __description__ + ' <' + __uri__ + '>'

__author__ = 'Shon T. Urbas, Geoff Johnson'
__email__ = 'shon.urbas@gmail.com, geoff.johnson@handshake.com'

__license__ = 'MIT'
__copyright__ = 'Copyright (c) 2016 Handshake Corp.'



import time

from freezegun import freeze_time
from mock import patch, call, Mock
import pytest
import boto3
from botocore.exceptions import ClientError

from pg2kinesis.stream import StreamWriter

@pytest.fixture()
def writer():
    with patch('aws_kinesis_agg.aggregator.RecordAggregator'), patch.object(boto3, 'client'):
        writer = StreamWriter('blah')
    return writer

def test__init__():
    mock_client = Mock()
    with patch.object(boto3, 'client', return_value=mock_client):
        error_response = {'Error': {'Code': 'ResourceInUseException'}}
        mock_client.create_stream = Mock(side_effect=ClientError(error_response, 'create_stream'))

        StreamWriter('blah')
        assert mock_client.create_stream.called
        assert call.get_waiter('stream_exists') in mock_client.method_calls, "We handled stream existence"

        error_response = {'Error': {'Code': 'Something else'}}
        mock_client.create_stream = Mock(side_effect=ClientError(error_response, 'create_stream'))

        mock_client.reset_mock()
        with pytest.raises(ClientError):
            StreamWriter('blah')
            assert mock_client.create_stream.called
            assert call.get_waiter('stream_exists') not in mock_client.method_calls, "never reached"


def test_put_message(writer):

    writer._send_agg_record = Mock()

    msg = Mock()
    msg.change.xid = 10
    msg.fmt_msg = object()

    writer.last_send = 1445444940.0 - 10      # "2015-10-21 16:28:50"
    with freeze_time('2015-10-21 16:29:00'):  # -> 1445444940.0
        result = writer.put_message(None)

        assert result is None, 'With no message or timeout we did not force a send'
        assert not writer._send_agg_record.called, 'we did not force a send'

        writer._record_agg.add_user_record = Mock(return_value=None)
        result = writer.put_message(msg)
        assert result is None, 'With message, no timeout and not a full agg we do not send'
        assert not writer._send_agg_record.called, 'we did not force a send'

    with freeze_time('2015-10-21 16:29:10'):  # -> 1445444950.0
        result = writer.put_message(None)
        assert result is not None, 'Timeout forced a send'
        assert writer._send_agg_record.called, 'We sent a record'
        assert writer.last_send == 1445444950.0, 'updated window'

    with freeze_time('2015-10-21 16:29:20'):  # -> 1445444960.0
        writer._send_agg_record.reset_mock()
        writer._record_agg.add_user_record = Mock(return_value='blue')
        result = writer.put_message(msg)

        assert result == 'blue', 'We passed in a message that forced the agg to report full'
        assert writer._send_agg_record.called, 'We sent a record'
        assert writer.last_send == 1445444960.0, 'updated window'


def test__send_agg_record(writer):
    assert writer._send_agg_record(None) is None, 'Do not do anything if agg_record is None'

    agg_rec = Mock()
    agg_rec.get_contents = Mock(return_value=(1, 2, 'datablob'))

    err = ClientError({'Error': {'Code': 'ProvisionedThroughputExceededException'}}, 'put_record')

    writer._kinesis.put_record = Mock(side_effect=[err, err, err, {'SequenceNumber': 12345}])

    with patch.object(time, 'sleep') as mock_sleep:
        writer._send_agg_record(agg_rec)
        assert mock_sleep.call_count == 3, "We had to back off 3 times so we slept"
        assert mock_sleep.call_args_list == [call(.1), call(.2), call(.4)], 'Geometric back off!'

    with pytest.raises(ClientError):
        writer._kinesis.put_record = Mock(side_effect=ClientError({'Error': {'Code': 'Something else'}},
                                                                  'put_record'))
        writer._send_agg_record(agg_rec)

    writer.back_off_limit = .3  # Will bust on third go around
    writer._kinesis.put_record = Mock(side_effect=[err, err, err, {'SequenceNumber': 12345}])
    with pytest.raises(Exception) as e_info, patch.object(time, 'sleep'):
        writer._send_agg_record(agg_rec)
        assert e_info.value.message == 'ProvisionedThroughputExceededException caused a backed off too many times!', \
            'We raise on too many throughput errors'



import codecs
import os
import re

from setuptools import setup, find_packages


###############################################################################

NAME = 'pg2kinesis'
PACKAGES = find_packages(where='.')

if not PACKAGES:
    raise RuntimeError('Package NOT FOUND!')

META_PATH = os.path.join('pg2kinesis', '__init__.py')
KEYWORDS = ['logical replication', 'kinesis', 'database']
CLASSIFIERS = [
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Natural Language :: English',
    'License :: OSI Approved :: MIT License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: Implementation :: CPython',
    'Topic :: Utilities',
]
INSTALL_REQUIRES = [
    'aws_kinesis_agg>=1.0.1',
    'boto3>=1.4.1',
    'botocore>=1.4.64',
    'click>=6.3.0',
    'protobuf>=3.0.0',
    'psycopg2>=2.7.1',
]

###############################################################################

HERE = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with codecs.open(os.path.join(HERE, *parts), 'rb', 'utf-8') as f:
        return f.read()


META_FILE = read(META_PATH)


def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = re.search(
        r'^__{meta}__ = [\'"]([^\'"]*)[\'"]'.format(meta=meta),
        META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError('Unable to find __{meta}__ string.'.format(meta=meta))


VERSION = find_meta('version')
URI = find_meta('uri')
LONG = read('README.rst')

if __name__ == "__main__":
    setup(
        name=NAME,
        description=find_meta("description"),
        license=find_meta("license"),
        url=URI,
        version=VERSION,
        author=find_meta("author"),
        author_email=find_meta("email"),
        maintainer=find_meta("author"),
        maintainer_email=find_meta("email"),
        keywords=KEYWORDS,
        long_description=LONG,
        packages=PACKAGES,
        zip_safe=True,
        classifiers=CLASSIFIERS,
        install_requires=INSTALL_REQUIRES,
        entry_points={
            'console_scripts': ['pg2kinesis=pg2kinesis.__main__:main'],
        }
    )



from __future__ import unicode_literals

from mock import Mock, call

from pg2kinesis.__main__ import Consume

def test_consume():
    mock_formatter = Mock(return_value='fmt_msg')
    mock_writer = Mock()

    consume = Consume(mock_formatter, mock_writer)

    mock_change = Mock()
    mock_change.data_start = 10
    mock_change.data_size = 100
    mock_change.payload = 'PAYLOAD'

    mock_writer.put_message = Mock(return_value=False)
    consume(mock_change)
    assert mock_writer.put_message.called, 'Sanity'
    assert call.cursor.send_feedback(flush_lsn=10) not in mock_change.mock_calls, \
        'we did not send feedback!'

    mock_writer.put_message = Mock(return_value=True)
    consume(mock_change)
    assert mock_writer.put_message.called, 'Sanity'
    assert call.cursor.send_feedback(flush_lsn=10) in mock_change.mock_calls, \
        'we sent feedback!'